30/11/97 11:50:43
   
bitsPerPixel: bitsPerPixel from: aStream
    "Returns an instance of the receiver capable of holding bitsPerPixel entries."

    ^self new
        bitsPerPixel: bitsPerPixel;
        fileInFrom: aStream;
        yourself!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isApplicationExtension! 
with: aGIFObject
    " Returns an instance of the receiver wrapping aGIFObject. "

    ^self new
        adaptee: aGIFObject;
        yourself!  
invalidObject: anObject
    " Private - The can't handle anObject. 
    Generates an error."

    ^anObject error: 'Don''t know how to handle this object.'!  
for: aGIFObject
    " Returns a proper adapter for aGIFObject. "

    | aClass |
    aClass := self allSubclasses reversed "IMPORTANT: in reverse order"
        detect: [:one| one canHandle: aGIFObject ]
        ifNone: [ self invalidObject: aGIFObject ].
    ^aClass with: aGIFObject!  
unknownElement: aGIFElement
    " aGIFElement is unknown.
    We can generate an error or returns nil for silenty ingore unknown elements. "

    self error: 'Unknown GIFElement'.

    ^nil "aGIFElement will be ignored."!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject.
    This method must be refined by concrete classes. "

    self subclasses size > 0 ifTrue: [ ^false ]. "abstract class"
    ^self implementedBySubclass!  
decompressorClass
	" Private - Returns the class for instances image-decompressors. "

	^GIFLZWDecompressor!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isImage!
for: aGIFElement
    " Returns a proper adapter for aGIFElement. "

    aGIFElement isUnknown
    ifTrue: [ ^self unknownElement: aGIFElement ].
    ^super for: aGIFElement!
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isControl!  
displayAt: aPoint with: aPen
    "Display the receiver at aPoint with: aPen"

    self client extension: self firstBlock!  
firstBlock
    " Private - Returns the first extension subblock."

    ^self adaptee subblocks first!  
adaptee
    "Returns the instance var of the receiver
     for external manipulation"
    ^adaptee!
adaptee: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    adaptee := anObject!   
client
    "Returns the instance var of the receiver
     for external manipulation"
    ^client!  
client: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    client := anObject! 
displayAt: aPoint with: aPen
    "Display the receiver at aPoint with: aPen
    By default we will do nothing.
    This method can be refined by subclasses."! 
isLoop
    "Returns true for a Netscape loop extension."

    ^false!  
screenDescriptor
    " Private - Returns the screen descriptor"

    ^self client screenDescriptor!
isValid
    " Returns true when the receiver can be used. "

    ^self adaptee notNil! 
colorMap
    " Private - Returns the color map."

    ^self client colorMap!   
token
    " Private - Returns the receiver's token. "

    ^16rFF! 
identifier
    " Private - Must returns an apropiate identifier for application. "

    ^self implementedBySubclass!   
canClone: anApplicationExtension
    " Private - Returns true if the receiver can clone anApplicationExtension.
    This method can be refined by subclasses. "

    ^self identifier = anApplicationExtension properties identifier! 
concreteApplicationExtensionFor: anApplicationExtension
    " Private - Returns an apropiate instance of the receiver that can be more concrete to handle contents of anApplicationExtension. "

    | aClass clone |
    aClass := self allSubclasses reversed "more concrete to abstract..."
        detect: [:one|
            clone := one cloneFor: anApplicationExtension.
            clone notNil
        ]
        ifNone: [ ^anApplicationExtension ].
    ^clone!
cloneFor: anApplicationExtension
    " Private - Returns an apropiate clone for anApplicationExtension. "

    (self canClone: anApplicationExtension) ifFalse: [ ^nil ].

    ^self new
        copyFrom: anApplicationExtension;
        yourself!   
blockSize
    " Returns the receiver's fixed block size in bytes. "

    ^11!  
token
    " Private - Returns the receiver's token.
    The concrete classes must refine this method. "

    self subclasses isEmpty
    ifTrue: [ self implementedBySubclass ].

    ^nil!
for: aToken
    " Returns an instance of the receievr capable of handling aToken."

    | aClass |
    aClass := self allSubclasses
        detect: [:one| one canHandle: aToken ]
        ifNone: [ self unknownGraphicControl ].
    ^aClass new
        token: aToken;
        yourself!  
from: aStream
    "Returns an instance of the receiver filled in from aStream."

    | aLabel graphicControl |
    aLabel := aStream nextByte.
    graphicControl := self for: aLabel.
    ^graphicControl fileInFrom: aStream
        "WARNING: This method MUST return the fileInFrom: result object."!  
unknownGraphicControl
    " Private - Returns the support for unKnown graphic controls. "

    ^GIFUnknownGraphicControl!  
canHandle: aToken
    " Private - Returns true when the receiver can handle aToken. "

    ^self token = aToken!   
from: aStream
    "Returns an instance of the receiver filled in from aStream."

    self implementedBySubclass!   
from: aStream in: aContainer
    "Returns an instance of the receiver filled in from aStream."

    | result |
    result := self from: aStream.
    result notNil ifTrue: [ result parent: aContainer ].
    ^result!  
subBlockClass
    " Private - Returns the support for sub-blocks.
    By default returns a generic sub-block.
    This method can be refined by subclasses. "

    ^GIFSubBlock! 
isNetscapeExtension
    "True if the receiver is aNetscapeExtension."

    ^false! 
copyContentsFrom: anApplicationExtension
    " Private - Fill the receiver's contents from anApplicationExtension.
    This method can be refined by subclasses."

    self contents: anApplicationExtension contents.!   
copyFrom: anApplicationExtension
    " Private - Fill the receiver's contents from anApplicationExtension.
    This method can be refined by subclasses."

    self
        token: anApplicationExtension token;
        copyContentsFrom: anApplicationExtension;
        yourself!   
adaptContent: bytes
    " Private - The receiver's can adapt bytes contents to internal representation."

    | aClass |
    aClass := bytes = self contents first
        ifTrue: [ GIFApplicationPropertiesSubBlock ]
        ifFalse: [ GIFApplicationDataSubBlock ].
    ^aClass
        in: self
        with: bytes!   
fileInFrom: aStream
    " Get the receiver's contents from aStream
    AND RETURNS the filled in container.
    Here we will detect proper aplication handlers."

    | anApplicationExtension clone |
    anApplicationExtension := super fileInFrom: aStream.
    anApplicationExtension isNil ifTrue: [ ^anApplicationExtension ].
    
    ^self species concreteApplicationExtensionFor: anApplicationExtension!
isApplicationExtension
    "True if the receiver is anApplicationExtension."

    ^true!   
properties
    " Returns the first property in the receiver's contents (or nil). "

    ^self contents detect: [:one| one isPropertySubBlock ] ifNone: []! 
subblocks
    " Returns contents without the properties. "

    | properties |
    properties := self properties.
    properties isNil ifTrue: [ ^self contents ].
    ^self contents copyWithout: properties

    !  
token
    "Returns the instance var of the receiver
     for external manipulation"
    ^token!
token: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    token := anObject!   
getContentsFrom: aStream
    " Private - Get the receiver's contents from aStream. "

    | blockSize block |
    contents := OrderedCollection new.
    [    blockSize := aStream nextByte.
        blockSize = 0 ] whileFalse: [
        block := self getNextBlock: blockSize from: aStream.
        contents add: block
    ].
    self adaptContents.! 
adaptContents
    " Private - The receiver's can adapt it contents to internal representation. "

    self contents: (self contents collect: [:each| self adaptContent: each])!
contents: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    contents := anObject! 
isComment
    "True if the receiver is aComment."

    ^false! 
isControl
    "True if the receiver is aControl."

    ^false! 
getNextBlock: blockSize from: aStream
    " Private - Returns the next block from aStream.
    By default the next blockSize characters will be read from aStream.
    This method can be refined by subclasses. "

    ^aStream next: blockSize!
contents
    "Returns the instance var of the receiver
     for external manipulation"
    ^contents!  
isUnknown
    "True if the receiver is unknown."

    ^false!  
isText
    "True if the receiver is Text."

    ^false!
isApplicationExtension
    "True if the receiver is anApplicationExtension."

    ^false!  
adaptContent: bytes
    " Private - The receiver's can adapt bytes contents to internal representation.
    The default action is to convert to subBlock elements.
    This method can be refined by subclasses."

    ^self species subBlockClass
        in: self
        with: bytes!   
fileInFrom: aStream
    " Get the receiver's contents from aStream
    AND RETURNS the fille in container.
    This method can be refined by subclasses. "

    self
        getContentsFrom: aStream;
        yourself!   
parent
    "Returns the instance var of the receiver
     for external manipulation"
    ^parent!  
parent: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    parent := anObject! 
in: aContainer with: bytes
    " Returns an instance of the receiver with bytes as its contents. "

    ^self new
        parent: aContainer;
        getContents: bytes;
        yourself! 
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isComment!  
displayAt: aPoint with: aPen
    "Display the receiver at aPoint with: aPen"

    "self text displayAt: aPoint + (0@20) with: aPen."
    ^super displayAt: aPoint with: aPen! 
text
    "Returns the receivers text."

    ^self firstBlock contents! 
with: aGIFSpecialElement
    " Returns an instance of extensionWrapper wrapping aGIFSpecialElement. "

    ^aGIFSpecialElement notImlpementedYet!  
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isSpecial!  
with: aGIFExtension
    " Returns an instance of extensionWrapper wrapping aGIFExtension. "

    ^GIFExtensionAdapter for: aGIFExtension extension!
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isExtension!
pathName: pathName
    " Returns an instance of the receiver read from file named pathName. "

    | stream result |
    stream := GIFDataStream pathName: pathName.
    result := self new sourceStream: stream.
    ^result!  
defaultControllerClass
    "Returns the default display controller class."

    ^GIFSilentController!  
fromStream: aStream
    "Returns an instance of the receiver obtained from aStream."

    | stream result |
    stream := GIFDataStream fromStream: aStream.
    result := self new sourceStream: stream.
    ^result!  
for: aGIFPlayer with: aGIFControlExtension
    "Returns an instance of the receievr initialized to control aGIFPlayer with aGIFControlExtension."

    ^self new
        client: aGIFPlayer;
        control: aGIFControlExtension;
        yourself!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^(super canHandle: aGIFObject)
    and: [ aGIFObject isNetscapeExtension ]!
isLoop
    "Returns true for a Netscape loop extension."

    ^self firstBlock isLoop! 
iterations
    "Returns true for a Netscape loop extension."

    ^self firstBlock iterations! 
version
    " Private - Must return the receiver's version string."

    ^'89a'!   
screenDescriptorClass
    " Private - Must return the class for supoprt of screen descriptor."

    ^GIF89aScreenDescriptor!   
version
    " Private - Must return the receiver's version string.
    This method must be refined by subclasses. "

    ^self implementedBySubclass! 
pathName: pathName
    "Returns an instance of the receiver to be read from file at pathName."

    ^self fromStream: (File pathNameReadOnly: pathName)!   
screenDescriptorClass
    " Private - Must return the class for support of screen descriptor.
    This method must be refined by subclasses. "

    ^self implementedBySubclass!  
checkVersionFrom: aStream
    " Private - By default we match stream contents with receiver's version.
    This method can be refined by subclasses. "

    | version |
    version := self version.
    version size = 0 ifTrue: [ ^true ].
    ^(aStream next: version size) = version!  
fromStream: aStream
    "Returns an instance of the receiver capable of handling stream aStream."

    | aClass |
    aClass := self allSubclasses
        detect: [:one| one canHandle: aStream ]
        ifNone: [ ^self invalidGIFFile ].
    ^aClass new
        readFrom: aStream;
        yourself!
colorMapClass
    " Private - Must return the class for support of global color tables.
    This method can be refined by subclasses. "

    ^GIFGlobalColorMap!  
canHandle: aStream
    " Must return true if the receiver can handle input from stream aStream. "

    | position result |
    position := aStream position.
    result := (aStream next: 3) = 'GIF'
        and: [ self checkVersionFrom: aStream  ].
    aStream position: position.
    ^result!   
invalidGIFFile
    " Private - Signal an error because the current file is invalid. "

    ^self error: 'Invalid GIF File.'!   
blockClass
    " Private - Must return the class supporting block elements.
    This method can be refined by subclasses. "

    ^GIFBlock!   
token
    " Private - Returns the receiver's token."

    ^16r2C!  
colorMapClass
    " Private - Must return the class for support of color tables.
    This method can be refined by subclasses. "

    ^GIFLocalColorMap!  
token
    " Private - Returns the receiver's token.
    This method must be refined by concrete classes. "

    self subclasses size = 0
    ifTrue: [ self implementedBySubclass ].

    ^nil!
for: aToken
    "Returns an instance of the receiver capable of handling byte token aToken."

    ^(self whoCanHandle: aToken) new
        token: aToken;
        yourself!  
defaultHandler
    " Private - Must return the default handler. "

    ^GIFUnknownBlock!   
for: aToken in: aContainer
    "Returns an instance of the receiver capable of handling byte token aToken."

    ^(self for: aToken)
        parent: aContainer;
        yourself!   
whoCanHandle: aToken
    " Private - For a concrete class the token will be used for test aToken.
    Else look in the receiver's inmediate subclasses for one that can handle aToken.
    if none returns the default handler. "

    ^self allSubclasses
        detect: [:one| one canHandle: aToken ]
        ifNone: [ self defaultHandler ]! 
canHandle: aToken
    " Private - Returns true if the receiver's token matches aToken. "

    ^self token = aToken!
extensionClass
    " Returns the root class for graphic controls support. "

    ^self implementedbySubclass!  
displayController
    "Returns the instance var of the receiver
     for external manipulation"

    displayController isNil ifTrue: [ self initializeDisplayController ].
    ^displayController!   
displayController: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    displayController := anObject!   
initializePlaying
    " Private - Initializes the receiver's playing"

    playing := false.!  
next
    "Returns the next element in the receiver's stream."

    | result |
    result := self stream next.
    (self playing and: [ self atEnd ])
    ifTrue: [
        self loopCount: self loopCount + 1.
        self isLooping ifTrue: [ self reset ].
    ].
    ^result!   
playAt: aPoint with: aPen
    "Start playing continuously with aPen until stoped or error ocurred (e.g. pen invalid)."

    [
		self playing: true.
		[ self displayAllAt: aPoint with: aPen ] on: Error do: [ self stop ].
	] forkAt: 3
        ! 
playing
    "Returns the instance var of the receiver
     for external manipulation"

    playing isNil ifTrue: [ self initializePlaying ].
    ^playing!   
boundingBox
    "Returns the receiver's boundingBox"

    ^0@0 extent: self extent!
playing: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    playing := anObject!   
displayAt: aPoint with: aPen
    "Display the receiver's current frame at aPoint with aPen.
    Does not advance the receiver."

    | current controller |
    current := self peek.
    controller := self displayController.
    controller prepareDisplay: current at: aPoint with: aPen.
    current displayAt: aPoint with: aPen.
    controller restoreDisplay: current at: aPoint with: aPen.!   
initializeDisplayController
    " Private - Initializes the receiver's displayController"

    displayController := self species defaultControllerClass new!   
colorMap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    colorMap := anObject! 
stream
    "Returns the receiver's streaming channel of frames.
	WARNING: This channel will be (re)built by demand."

	| collection element location |
	sourceStream notNil ifTrue: [ "more elements to come..."
		sourceStream atEnd ifTrue: [ sourceStream := nil. ^self stream ].
		(stream isNil or: [ stream atEnd ]) "one element is needed. NOW"
		ifTrue: [
			collection := stream isNil
				ifTrue: [ location := 0. OrderedCollection new ]
				ifFalse: [ location := stream position. stream contents ].
			element := self adapt: sourceStream next.
			(element notNil and: [ element isValid ])
			ifTrue: [ collection addLast: element].
			stream := ReadStream on: collection.
			stream position: location.
			^self stream "lets (re)built again if needed"
		].
	].
    ^stream!   
iterations
    "Returns the instance var of the receiver
     for external manipulation"

    iterations isNil ifTrue: [ self initializeIterations ].
    ^iterations!   
displayAllAt: aPoint with: aPen
    "Display all the receiver's frames at aPoint with aPen."

    self reset.
    [self atEnd] whileFalse: [
        self displayAt: aPoint with: aPen.
        self next
    ].!  
extent
    "Returns the receiver's extent"

    ^self screenDescriptor extent! 
backgroundColor
    "Returns the receiver's background color."

    ^self colorMap colorAt: self screenDescriptor background!  
isLooping
    "Returns true if the receiver has a Netscape loop control and must loop."

    ^self iterations = 0 "infinite iterations"
    or: [ self loopCount < self iterations ]! 
peek
    "Peek the next element in the receiver's stream."

    ^self stream peek! 
iterations: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    iterations := anObject! 
initializeIterations
    " Private - Initializes the receiver's iterations"

    | contents |
    contents := self stream contents.
    iterations := (contents notEmpty and: [ contents first isLoop ])
        ifTrue: [ contents first iterations ]
        ifFalse: [ -1 ].!   
screenDescriptor: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    screenDescriptor := anObject! 
sourceStream: aGIFDataStream
    " Private - Sets the sourceStream and related instances in the receiver. "

    sourceStream := aGIFDataStream.
	self
		screenDescriptor: sourceStream screenDescriptor;
		colorMap: sourceStream colorMap;
		yourself.!  
adapt: aGIFElement
    " Private - Adapt aGIFElement to an element usefull to the receiver. "

    | adapter |
    aGIFElement isNil ifTrue: [ ^nil ].
    adapter := GIFElementAdapter for: aGIFElement.
    adapter notNil ifTrue: [ adapter client: self ].
    ^adapter!   
displayAllAt: aPoint
    "Display all the receiver's frames at aPoint with Display pen."

    ^self displayAllAt: aPoint with: Display pen!
transparentColor: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    transparentColor := anObject! 
transparentColor
    "Returns the instance var of the receiver
     for external manipulation"
    ^transparentColor!  
colorMap
    "Returns the instance var of the receiver
     for external manipulation"
    ^colorMap!  
stop
    "Stop playing."

    self playing: false.!
atEnd
    " Returns true when the receiver is atEnd. "

    ^self stream atEnd!
screenDescriptor
    "Returns the instance var of the receiver
     for external manipulation"
    ^screenDescriptor!  
loopCount
    "Returns the instance var of the receiver
     for external manipulation"

    loopCount isNil ifTrue: [ self initializeLoopCount ].
    ^loopCount!   
loopCount: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    loopCount := anObject!   
extension: aGIFControlExtension
    "Sets the controller to aGIFControlExtension."

    self displayController:
        (GIFImageController
            for: self
            with: aGIFControlExtension)!  
initializeLoopCount
    " Private - Initializes the receiver's loopCount"

    loopCount := 0.!
reset
    "Reset the receiver to start position."

    self stream reset!  
interlace
    "Returns true if the receiver is codified as interlaced."

    ^self descriptor interlace!   
bitmap
    "Returns the instance var of the receiver
     for external manipulation"

    bitmap isNil ifTrue: [ self initializeBitmap ].
    ^bitmap!   
bitsPerPixel
    "Returns the receiver's bitsPerPixel."

    ^self client screenDescriptor bitsPerPixel!   
displayAt: aPoint with: aPen
    "Display the receiver at aPoint with: aPen"

    | aBitmap box |
    aBitmap := self bitmap.
    aBitmap notNil
    ifTrue: [
        aBitmap
            displayAt: aPoint + self leftTop
            with: aPen
    ]
    ifFalse: [
        box := aPoint + self leftTop extent: self extent.
        aPen
            place: box leftBottom;
            goto: box rightTop;
            place: box leftTop;
            box: box rightBottom;
            goto: box leftTop;
            place: box rightBottom;
            yourself
    ]!   
decompressor
    " Private - Returns an apropiate image-decompressor for the receiver."

	^self species decompressorClass for: self!   
extent
    "Returns the receiver's extent."

    ^self descriptor extent!  
descriptor
    "Returns the image descriptor."

    ^self adaptee! 
initializeBitmap
    " Private - Initializes the receiver's bitmap"

    bitmap := self decompressor getBitmap.!   
chunks
    "Returns the receiver's information chunks."

    ^self adaptee imageData extension contents
		collect: [:each| each contents ]!   
transparentColor
    "Returns the image transparentColor."

    ^self client transparentColor! 
bitmap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    bitmap := anObject! 
leftTop
    "Returns the receiver's leftTop."

    ^self descriptor leftTop!   
chunkStream
    "Returns aStream with the receiver's information chunks."

    ^ReadStream on: self chunks!
isDataSubBlock
    " Returns true for data sub-blocks. "

    ^true!   
contents
    "Returns the instance var of the receiver
     for external manipulation"
    ^contents!  
isDataSubBlock
    " Returns true for data sub-blocks. "

    ^false!  
contents: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    contents := anObject! 
isControlSubBlock
    " Returns true for control sub-blocks. "

    ^false!
isPropertySubBlock
    " Returns true for property sub-blocks. "

    ^false!  
copyFrom: aDataSubBlock
    " Private - Copy the receiver's contents from aDataSubBlock. "

    self
        parent: aDataSubBlock parent;
        contents: aDataSubBlock contents! 
getContents: bytes
    " Private - Fill the receiver contents with bytes.
    This method can be refined by subclasses for decription of bytes."

    self contents: bytes.!  
isImage
    "Returns true if the receiver is a GraphicImageDescriptor block"

    ^false!  
isControl
    "Returns true if the receiver is a Control block"

    ^false!   
isSpecial
    "Returns true if the receiver is a SpecialRendering block"

    ^false!  
isTrailer
    "Returns true if the receiver is a Trailer block"

    ^false!   
token
    "Returns the instance var of the receiver
     for external manipulation"
    ^token!
isUnknown
    "Returns true if the receiver is a Unknown block"

    ^false!   
token: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    token := anObject!   
isExtension
    "Returns true if the receiver is a GraphicExtension block"

    ^false!
fileInFrom: aStream
    " Get the receiver's contents from aStream. "

    self implementedBySubclass.!
isGraphic
    "Returns true if the receiver is a GraphicRendering block"

    ^false!  
isPropertySubBlock
    " Returns true for property sub-blocks. "

    ^true!   
token
    " Private - Returns the receiver's token.
    The unKnown handlers has undefined token. "

    ^nil!
isControl
    "Returns true if the receiver is a Control block"

    self token = 16r3B ifTrue: [ ^true ].
    ^self token between: 16r80 and: 16rF9! 
isUnknown
    "Returns true if the receiver is a Unknown block"

    ^true!
isGraphic
    "Returns true if the receiver is a GraphicRendering block"

    self token = 16r3B ifTrue: [ ^false ].
    ^self token between: 16r00 and: 16r7F!   
fileInFrom: aStream
    " Get the receiver's contents from aStream. "

    ^self notImplementedYet!
isSpecial
    "Returns true if the receiver is a SpecialRendering block"

    ^self token between: 16rFA and: 16rFF!   
for: aGIFGraphicImageAdapter
	"Returns an instance of the receiver capable of decompression image of aGIFGraphicImageAdapter"

	^self new
		initializeWith: aGIFGraphicImageAdapter;
		yourself! 
transparentImageClass
	" Private - Returns the class for Transparent bitmapping support. "

	^Smalltalk at: #TransparentImage ifAbsent: [ "unsupported feature" ]! 
prepareDisplay: aGIFAdapter at: aPoint with: aPen
    "The receiver can prepare the medium to display aGIFAdapter at aPoint with aPen.
    The default action is to do nothing.
    This method can be implemented by subclasses"! 
restoreDisplay: aGIFAdapter at: aPoint with: aPen
    "The receiver can restore the medium from display aGIFAdapter at aPoint with aPen.
    The default action is to do nothing.
    This method can be implemented by subclasses"!   
control
    "Returns the instance var of the receiver
     for external manipulation"
    ^control!
control: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    control := anObject!   
client: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    client := anObject! 
client
    "Returns the instance var of the receiver
     for external manipulation"
    ^client!  
token
    " Private - Returns the receiver's token."

    ^16r21!  
extensionClass
    " Returns the root class for graphic controls support. "

    ^GIFGraphicExtensionContents! 
isExtension
    "Returns true if the receiver is a GraphicExtension block"

    ^true! 
extension
    "Returns the instance var of the receiver
     for external manipulation"
    ^extension!
extension: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    extension := anObject!   
fileInFrom: aStream
    " Get the receiver's contents from aStream. "

    self extension: (self species extensionClass from: aStream in: self)!   
isGraphic
    "Returns true if the receiver is a GraphicRendering block"

    ^true!   
token
    " Private - Returns the sub-block token. "

    ^16r01!  
token
    " Private - Returns the sub-block token. 
    This method must be implemented in concrete classes."

    self subclasses size = 0 ifTrue: [
        ^self implementedBySubclass
    ].

    ^nil!   
from: aDataSubBlock
    " Returns an instance of the receiver formed from aDataSubBlock. "

    | token aClass |
    token := (ReadWriteStream on: aDataSubBlock contents) nextByte.
    aClass := self allSubclasses reversed
        detect: [:one | one token = token ]
        ifNone: [ self ].
    ^aClass new copyFrom: aDataSubBlock! 
extensionClass
    " Returns the root class for graphic controls support. "

    ^GIFRasterDataBlocks! 
validTokenInterval
    " Private - Returns the valid tokens to be handled by receiver's instances. "

    ^2 to: 8!
canHandle: aToken
    " Private - aToken can be managed if is in receiver's token interval. "

    ^self validTokenInterval includes: aToken!  
minimumCodeSize
    "Returns the Minimum Code Size.
    This byte determines the initial number of bits used for LZW codes in the image data, as described in Appendix F."

    ^self token!  
token
    "Returns the instance var of the receiver
     for external manipulation"
    ^token!
token: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    token := anObject!   
isLoop
    "Returns true for a Loop block"

    ^false!
copyFrom: aDataSubBlock
    " Private - Copy the receiver's contents from aDataSubBlock. "

    super copyFrom: aDataSubBlock.
    self fileInFrom: (ReadWriteStream on: self contents).! 
fileInFrom: aStream
    " Private - Fill the receiver's contents from aStream.
    By default only fill the receiver's token, but this method can be completed by subclasses. "

    self
        token: aStream nextByte;
        yourself!
isControlSubBlock
    " Returns true for control sub-blocks. "

    ^true! 
from: aStream
    "Returns an instance of the receiver filled in from aStream."

    ^self new fileInFrom: aStream
        "WARNING: This method MUST return the fileInFrom: result object."! 
subBlockClass
    " Private - Returns the support for sub-blocks. "

    ^GIFLZWCompressedSubBlock!
size
	"Returns the size of the receiver."

	^self entries size!
bitsPerPixel: bitsPerPixel
    " Private - Prepare the receievr to hold bitsPerPixel entries. "

    entries := Array new: (1 timesTwoPower: bitsPerPixel) truncated.! 
entries
    "Returns the instance var of the receiver
     for external manipulation"
    ^entries!
fileInFrom: aStream
    "Get the receiver's contents from aStream."

    | color |
    1 to: self entries size do: [:i|
        color := Color
			red: aStream nextByte
			green: aStream nextByte
			blue: aStream nextByte.
        self entries at: i put: color.
    ]! 
colorAt: index
    "Return the color at index <index> (or nil).
    WARNING: index is ZERO indexed."

    (index + 1 between: 1 and: self entries size) ifFalse: [ ^nil ].
    ^self entries at: index + 1!  
prepareDisplay: aGIFAdapter at: aPoint with: aPen
    "The receiver can prepare the medium to display aGIFAdapter at aPoint with aPen."

    self control disposeSelector = #previous
    ifTrue: [
        self buffer pen
            copy: aPen
            from: (self client boundingBox translateBy: aPoint)
            at: 0@0
    ].

    self considerTransparency: aGIFAdapter at: aPoint with: aPen.!  
restoreDisplay: aGIFAdapter at: aPoint with: aPen
    "The receiver can restore the medium from display aGIFAdapter at aPoint with aPen."

    self
        considerUserInput: aGIFAdapter at: aPoint with: aPen;
        considerDelayTime: aGIFAdapter at: aPoint with: aPen;
        disposeDisplay: aGIFAdapter at: aPoint with: aPen;
        yourself.

    self client displayController: nil. "free the controler"!  
waitAction: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    waitAction := anObject! 
disposeDisplay: aGIFAdapter at: aPoint with: aPen
    " Private - Consider dispose method to restore display."

    | selector |
    selector := self control disposeSelector.
    selector = #none ifTrue: [ ^self ].

    selector = #background
    ifTrue: [
        aPen fill: self client boundingBox color: self client backgroundColor.
        ^self
    ].

    selector = #previous
    ifTrue: [
        aPen
            copy: self buffer pen
            from: self buffer boundingBox
            at: aPoint.
        ^self
    ]!
considerDelayTime: aGIFAdapter at: aPoint with: aPen
    " Private - Consider the delayTime value. "

    | deadTime delayTime |
    delayTime := control delayTime.
    delayTime > 0 ifFalse: [ ^self ].

    deadTime := delayTime * 10 + Time millisecondClockValue.
    [ Time millisecondClockValue > deadTime ]
    whileFalse: [ self waitAction evaluate ].!
considerTransparency: aGIFAdapter at: aPoint with: aPen
    " Private - Consider the transparency index settings value. "

    control useTransparentColor ifFalse: [ ^self ].
    self client transparentColor:
        (self client colorMap colorAt: self control transparentColorIndex)! 
considerUserInput: aGIFAdapter at: aPoint with: aPen
    " Private - Consider the user input flag. "

    control needsUserInput ifFalse: [ ^self ].

    "user input will be ingnored."!
buffer
    "Returns the instance var of the receiver
     for external manipulation"

    buffer isNil ifTrue: [ self initializeBuffer ].
    ^buffer!   
buffer: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    buffer := anObject! 
waitAction
    "Returns the instance var of the receiver
     for external manipulation"
    ^waitAction!  
initializeBuffer
    " Private - Initializes the receiver's buffer"

    buffer := Bitmap screenExtent: self client extent.!   
getExtrasFrom: aStream
    " Private - Get the receiver's extra information color from aStream."

    ^aStream nextByte.!  
bitsPerPixel
    "Returns the instance var of the receiver
     for external manipulation"
    ^bitsPerPixel!  
background: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    background := anObject! 
bitsPerPixel: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    bitsPerPixel := anObject! 
getExtentFrom: aStream
    " Private - Get the receiver's extent from aStream."

    | width height |
    width := aStream nextUShort.
    height := aStream nextUShort.
    ^self extent: width @ height!  
colorResolution: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    colorResolution := anObject!   
colorResolution
    "Returns the instance var of the receiver
     for external manipulation"
    ^colorResolution!
sortOrder
    "Returns the instance var of the receiver
     for external manipulation"
    ^sortOrder!
sortOrder: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    sortOrder := anObject!   
extent
    "Returns the instance var of the receiver
     for external manipulation"
    ^extent!  
extent: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    extent := anObject! 
background
    "Returns the instance var of the receiver
     for external manipulation"
    ^background!  
getFlagsFrom: aStream
    " Private - Get the receiver's flags from aStream."

    | flags |
    flags := aStream nextByte.
    self
        useColorMap: (flags bitAnd: 2r10000000) ~= 0;
        colorResolution: ((flags bitAnd: 2r01110000) bitShift: -4) + 1;
        sortOrder:  (flags bitAnd: 2r00001000) ~= 0;
        bitsPerPixel: (flags bitAnd: 2r00000111) + 1;
        yourself.
    ^flags!
useColorMap
    "Returns the instance var of the receiver
     for external manipulation"
    ^useColorMap!
useColorMap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    useColorMap := anObject!   
fileInFrom: aStream
    "Get the receiver's contents from aStream."

    self
        getExtentFrom: aStream;
        getFlagsFrom: aStream;
        getBackgroundFrom: aStream;
        getExtrasFrom: aStream;
        yourself!
getBackgroundFrom: aStream
    " Private - Get the receiver's background color from aStream."

    | byte |
    byte := aStream nextByte.
    self background: byte.
    ^byte! 
token
    " Private - Returns the receiver's token."

    ^16r3B!  
isTrailer
    "Returns true if the receiver is a Trailer block"

    ^true!
fileInFrom: aStream
    " Get the receiver's contents from aStream. "

    "empty element."!   
isControl
    "Returns true if the receiver is a control block"

    ^true!
token
    " Private - Returns the receiver's token. "

    ^16r01! 
adaptContent: bytes
    " Private - The receiver's can adapt bytes contents to internal representation."

    | aClass |
    aClass := bytes = self contents first
        ifTrue: [ GIFPlainTextPropertiesSubBlock ]
        ifFalse: [ GIFPlainTextSubBlock ].
    ^aClass
        in: self
        with: bytes!   
isText
    "True if the receiver is Text."

    ^true! 
token
    " Private - Returns the receiver's token. "

    ^16rFE! 
subBlockClass
    " Private - Returns the support for sub-blocks."

    ^GIFDataSubBlock!  
bitsPerPixel
    "Returns the instance var of the receiver
     for external manipulation"
    ^bitsPerPixel!  
colorTableFor: bitCount
    " Private - Returns a byte array with color table compatible with Windows colorTable obtained from the receiver.
    Or nil if the receiver has no color table."

    | stream colorMap color bytes paletteSize |
    bytes := ByteArray new: (WinBitmapInfo colorTableSize: bitCount).
    stream := WriteStream on: bytes.
    colorMap := self colorMap.
    colorMap isNil ifTrue: [ ^nil ].
    1 to: colorMap size do: [:index|
		color := colorMap colorAt: index - 1.
        stream
            nextPut: color blue;
            nextPut: color green;
            nextPut: color red;
            nextPut: 0;
            yourself
    ].
    ^bytes! 
getDIB
    " Private - Get a DIB from the receiver's internal information. "

    | outCodes outCount bitMask code curCode oldCode inCode finChar i
	bytes dib bitmapInfo |

    outCodes := ByteArray new: 1025.
    outCount := 0.
    bitMask := (1 bitShift: self bitCount "bitsPerPixel") - 1.
	self initializeCompressionTables.

    self setParameters: self initCodeSize.
    bytes := ByteArray new: self rowByteSize * self height.

    [ (code := self nextBits) = eoiCode ]
	whileFalse: [
		code = clearCode
        ifTrue: [
			self setParameters: self initCodeSize.
			curCode := oldCode := code := self nextBits.
			finChar := curCode bitAnd: bitMask.
			"Horrible hack to avoid running off the end of the bitmap.
			 Seems to cure problem reading some gifs!!? tk 6/24/97 20:16"
			xpos = 0
			ifTrue: [
            	ypos < self height
				ifTrue: [
					bytes
						at: ypos * self rowByteSize + xpos + 1
						put: finChar
				]
			]
			ifFalse: [
				bytes
					at: ypos * self rowByteSize + xpos + 1
					put: finChar
			].
			self updatePixelPosition
		]
		ifFalse: [
			curCode := inCode := code.
			curCode >= freeCode
			ifTrue: [
				curCode := oldCode.
				outCodes at: (outCount := outCount + 1) put: finChar
			].
			[curCode > bitMask] whileTrue: [
				outCount > 1024
				ifTrue: [ ^self error: 'corrupt GIF file (OutCount)' ].
				outCodes
					at: (outCount := outCount + 1)
					put: (suffixTable at: curCode + 1).
				curCode := prefixTable at: curCode + 1
			].
			finChar := curCode bitAnd: bitMask.
			outCodes at: (outCount := outCount + 1) put: finChar.
			i := outCount.
			[ i > 0 ] whileTrue: [
				bytes at: ypos * self rowByteSize + xpos + 1 put: (outCodes at: i).
				self updatePixelPosition.
				i := i - 1
			].
			outCount := 0.
			prefixTable at: freeCode + 1 put: oldCode.
			suffixTable at: freeCode + 1 put: finChar.
			oldCode := inCode.
			freeCode := freeCode + 1.
			self checkCodeSize
		]
	].

	self clearCompressionTables.

    bitmapInfo := self
        newBitmapInfoWidth: self width
        height: self height
        bitCount: (self bitCount max: 8).
    dib := DIB new
        bitmapInfo: bitmapInfo;
        bits: bytes;
        yourself.
	dib bits: (self reversedBitsOf: dib).
    ^dib
!  
bitsPerPixel: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    bitsPerPixel := anObject! 
checkBuffer
	" Private - Check if we need to fill the buffer.
	Retursn true if successful, false if failed (e.g. atEnd). "

    ^bufStream atEnd not or: [ self canFillBuffer ]!  
getBitmap
    " Private - Get a new bitmap from the receiver's internal information. "

	| dib transparentImageClass |
	dib := self getDIB.
	transparentImageClass := self transparentImageClass.
    ^(self transparentColor isNil or: [ transparentImageClass isNil])
	ifTrue: [ dib asBitmap ]
	ifFalse: [
		transparentImageClass
			fromDIB: dib
			transparentColor: self transparentColor asInteger
	]!
bitCount
    "Returns the receiver's bitCount."

    | bitCount |
    bitCount := self bitsPerPixel * self rowByteSize // self width.

    "Make sure bits per pixel is valid"
    ^bitCount <= 1 ifTrue: [ 1 ]
        ifFalse: [ bitCount <= 4 ifTrue: [ 4 ]
        ifFalse: [ bitCount <= 8 ifTrue: [ 8 ]
        ifFalse: [ bitCount <= 24 ifTrue: [ 24 ]
        ifFalse: [ 4 ]]]
    ]!
rowByteSize
    "Returns the instance var of the receiver
     for external manipulation"
    ^rowByteSize!
rowByteSize: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    rowByteSize := anObject!   
peekByte
	" Private - Peek the next available byte from the receiver's input or nil if atEnd. "

   ^self checkBuffer ifTrue: [ bufStream peek ]!  
reversedBitsOf: aDIB
    "Private - Revert the image bits of aDIB."

    ^self
		flip: aDIB bits copy
		by: (self scanLineSizeOf: aDIB)
!   
scanLineSizeOf: aDIB
    "Private - Rturns the scan line size of aDIB."

    | bitmapInfo |
    bitmapInfo := aDIB bitmapInfo.
    ^bitmapInfo sizeImage = 0
    ifTrue: [ ((bitmapInfo width * bitmapInfo bitCount) + 31) // 32 * 4 ]
    ifFalse: [ bitmapInfo sizeImage // bitmapInfo height ]! 
updatePixelPosition
	" Private - Update cursor for pixel offset. "

    (xpos := xpos + 1) >= self width ifFalse: [ ^self ].
    xpos := 0.
    interlace ifFalse: [ ypos := ypos + 1. ^self ].
    pass = 0 ifTrue: [
		(ypos := ypos + 8) >= self height
		ifTrue: [ pass := pass + 1. ypos := 4 ].
		^self
	].
    pass = 1 ifTrue: [
		(ypos := ypos + 8) >= self height
        ifTrue: [ pass := pass + 1. ypos := 2].
        ^self
	].
    pass = 2 ifTrue: [
		(ypos := ypos + 4) >= self height
		ifTrue: [ pass := pass + 1. ypos := 1 ].
        ^self
	].
    pass = 3 ifTrue: [ ypos := ypos + 2. ^self ].

    ^self error: 'can''t happen'!
colorTable
    " Private - Returns a byte array with color table compatible with Windows colorTable obtained from the receiver.
    Or nil if the receiver has no color table."

    ^self colorTableFor: self bitCount!  
initializeWith: aGIFGraphicImageAdapter
	" Private - Initialize the receiver for decompression of image in aGIFGraphicImageAdapter. "

	self
		descriptor: aGIFGraphicImageAdapter descriptor;
		bitsPerPixel: aGIFGraphicImageAdapter bitsPerPixel;
		chunkStream: aGIFGraphicImageAdapter chunkStream;
		interlace: aGIFGraphicImageAdapter interlace;
		transparentColor: aGIFGraphicImageAdapter transparentColor;
		yourself.
    self bitsPerPixel > 8 ifTrue: [ ^self error: 'Never heard of a GIF that deep' ].
    pass := xpos := ypos := 0.
    remainBitCount := 0.
    self rowByteSize: self width + 3 // 4 * 4.
    bufStream := ReadStream on: ByteArray new.
	self
		initializeCompressionTables;
		yourself.
!
newBitmapInfoWidth: width height: height bitCount: bitCount
    " Private - Returns WinBitmapInfoHeader initialized to be part of a DIB of width@height compatible with the receiver."

    | bitmapInfo colorTable |

    bitmapInfo := WinBitmapInfo
        width: width height: height
        planes: 1 bitCount: bitCount.

   bitmapInfo
        compression: 0; "BI_RGB"
        sizeImage: bitmapInfo bitCount * width + 31 // 32 * 4 * height;
        xPelsPerMeter: 2834;
        yPelsPerMeter: 2834;
        clrUsed: 0;
        clrImportant: 0;
        yourself.

    (colorTable := self colorTable) notNil
    ifTrue: [ bitmapInfo colorTable: colorTable ]
    ifFalse: [
        bitmapInfo colorTable: #(0 0 0 0 255 255 255 0)
    ].

    ^bitmapInfo!   
checkCodeSize
	" Private - Check if we need to grow code tables. "

    (freeCode > maxCode and: [ codeSize < 12 ])
    ifTrue: [
        codeSize := codeSize + 1.
        maxCode := (1 bitShift: codeSize) - 1
    ].!  
canFillBuffer
	" Private - Returns true if the receiver can re-fill the input buffer. "

	self chunkStream atEnd ifTrue: [ ^false ].
    bufStream := ReadStream on: self chunkStream next asByteArray.
	^true!  
transparentImageClass
	" Private - Returns the class for Transparent bitmapping support. "

	^self species transparentImageClass!  
flip: bits by: scanLineSize
    " Private - Flip the vertically the bits by scanLineSize."

    | buffer j |
    j := bits size.
    buffer := ByteArray new: scanLineSize.
    1 to: bits size - scanLineSize // 2 by: scanLineSize do: [:i|
        j := j - scanLineSize.
        buffer replaceFrom: 1 to: scanLineSize with: bits startingAt: i.
        bits
            replaceFrom: i to: i + scanLineSize - 1 with: bits startingAt: j + 1;
            replaceFrom: j + 1 to: j + scanLineSize with: buffer startingAt: 1
    ].
	^bits!   
nextBits
	" Private - Get the next available bits from the receiver's input. "

    | integer readBitCount shiftCount byte |
    integer := 0.
    remainBitCount = 0
	ifTrue: [
        readBitCount := 8.
        shiftCount := 0
    ]
    ifFalse: [
        readBitCount := remainBitCount.
        shiftCount := remainBitCount - 8
    ].
    [readBitCount < codeSize] whileTrue: [
        byte := self nextByte.
        byte == nil ifTrue: [ ^eoiCode ].
        integer := integer + (byte bitShift: shiftCount).
        shiftCount := shiftCount + 8.
        readBitCount := readBitCount + 8
    ].

    (remainBitCount := readBitCount - codeSize) = 0
    ifTrue: [ byte := self nextByte ]
    ifFalse: [ byte := self peekByte ].

    byte == nil ifTrue: [ ^eoiCode ].
    ^(integer + (byte bitShift: shiftCount)) bitAnd: maxCode!   
nextByte
	" Private - Get the next available byte from the receiver's input or nil if atEnd. "

   ^self checkBuffer ifTrue: [ bufStream next ]!   
chunkStream: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    chunkStream := anObject!   
interlace
    "Returns the instance var of the receiver
     for external manipulation"
    ^interlace!
interlace: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    interlace := anObject!   
initCodeSize
    "Returns the receiver's initCodeSize."

    ^self descriptor imageData minimumCodeSize!   
clearCompressionTables
	" Private - Clear the receiver's compression tables. "

    prefixTable := suffixTable := nil.!
width
    "Returns the receiver's width."

    ^self extent x! 
height
    "Returns the receiver's height."

    ^self extent y!   
extent
    "Returns the receiver's extent."

    ^self descriptor extent!  
initializeCompressionTables
	" Private - Initialize the receiver's compression tables. "

    prefixTable := Array new: 4096.
    suffixTable := Array new: 4096.!
descriptor
    "Returns the instance var of the receiver
     for external manipulation"
    ^descriptor!  
setParameters: initCodeSize
    " Private - Sets the receiver's parameters for [de]compression operations."

    clearCode := 1 bitShift: initCodeSize.
    eoiCode := clearCode + 1.
    freeCode := clearCode + 2.
    codeSize := self initCodeSize + 1. 
    maxCode := (1 bitShift: codeSize) - 1!
descriptor: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    descriptor := anObject! 
transparentColor: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    transparentColor := anObject! 
transparentColor
    "Returns the instance var of the receiver
     for external manipulation"
    ^transparentColor!  
colorMap
    "Returns the receiver's colorMap."

	| map |
    map := self descriptor colorMap.
	map isNil ifTrue: [
		map := self descriptor globalColorMap
	].
	^map!
chunkStream
    "Returns the instance var of the receiver
     for external manipulation"
    ^chunkStream!
identifier
    " Private - Must returns an apropiate identifier for application. "

    ^'NETSCAPE'!   
isNetscapeExtension
    "True if the receiver is aNetscapeExtension."

    ^true!  
copyContentsFrom: anApplicationExtension
    " Private - Fill the receiver's contents from anApplicationExtension. "

    | collection |
    collection := anApplicationExtension contents.
    collection := collection collect: [:each|
        each isDataSubBlock
        ifTrue: [ GIFNetscapeDataSubBlock from: each ]
        ifFalse: [ each ]
    ].
    self contents: collection.!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isTrailer!  
isSpecial
    "Returns true if the receiver is a SpecialRendering block"

    ^true!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isText! 
getExtrasFrom: aStream
    " Private - Get the receiver's extra information color from aStream."

    "do nothing here..."!
interlace
    "Returns the instance var of the receiver
     for external manipulation"
    ^interlace!
interlace: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    interlace := anObject!   
globalColorMap
    "Returns the global color map."

    ^self parent colorMap! 
colorMap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    colorMap := anObject! 
getExtentFrom: aStream
    " Private - Get the receiver's extent from aStream. "

    | width height |
    width := aStream nextUShort.
    height := aStream nextUShort.
    ^self extent: width @ height! 
imageData
    "Returns the instance var of the receiver
     for external manipulation"
    ^imageData!
imageData: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    imageData := anObject!   
sort
    "Returns the instance var of the receiver
     for external manipulation"
    ^sort!  
extent
    "Returns the instance var of the receiver
     for external manipulation"
    ^extent!  
reserved
    "Returns the instance var of the receiver
     for external manipulation"
    ^reserved!  
getImageDataFrom: aStream
    " Private - Get the receiver's image data from aStream."

    "Image data must be contiguous in parent data stream."
    self imageData: self parent next.! 
getLeftTopFrom: aStream
    " Private - Get the receiver's leftTop from aStream. "

    | left top |
    left := aStream nextUShort.
    top := aStream nextUShort.
    ^self leftTop: left @ top!  
extent: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    extent := anObject! 
reserved: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    reserved := anObject! 
sort: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    sort := anObject! 
getFlagsFrom: aStream
    " Private - Get the receiver's flags from aStream."

    | flags |
    flags := aStream nextByte.
    self
        useColorMap: (flags bitAnd: 2r10000000) ~= 0;
        interlace: (flags bitAnd: 2r01000000) ~= 0;
        sort:  (flags bitAnd: 2r00100000) ~= 0;
        reserved: ((flags bitAnd: 2r00011000) bitShift: -3);
        sizeOfLocalColorTable: (flags bitAnd: 2r00000111);
        yourself.
    ^flags!  
useColorMap
    "Returns the instance var of the receiver
     for external manipulation"
    ^useColorMap!
useColorMap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    useColorMap := anObject!   
isImage
    "Returns true if the receiver is a GraphicImageDescriptor block"

    ^true!   
colorMap
    "Returns the instance var of the receiver
     for external manipulation"
    ^colorMap!  
sizeOfLocalColorTable
    "Returns the instance var of the receiver
     for external manipulation"
    ^sizeOfLocalColorTable!
sizeOfLocalColorTable: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    sizeOfLocalColorTable := anObject!   
leftTop
    "Returns the instance var of the receiver
     for external manipulation"
    ^leftTop!
leftTop: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    leftTop := anObject!   
fileInFrom: aStream
    " Get the receiver's contents from aStream. "

    self
        getLeftTopFrom: aStream;
        getExtentFrom: aStream;
        getFlagsFrom: aStream;
        getExtrasFrom: aStream;
        getColorMapFrom: aStream;
        getImageDataFrom: aStream;
        yourself!  
getColorMapFrom: aStream
    " Private - Get the receiver's colorMap information color from aStream."

    | map |
    self useColorMap ifFalse: [ ^self colorMap: nil ].
    map := self species colorMapClass
        bitsPerPixel: self bitsPerPixel from: aStream.
    self colorMap: map.!
displayAt: aPoint with: aPen
    "Display the receiver at aPoint with: aPen"

    self notImplementedYet.
"
25. Plain Text Extension.

      a. Description. The Plain Text Extension contains textual data and the
      parameters necessary to render that data as a graphic, in a simple form.
      The textual data will be encoded with the 7-bit printable ASCII
      characters.  Text data are rendered using a grid of character cells
      defined by the parameters in the block fields. Each character is rendered
      in an individual cell. The textual data in this block is to be rendered
      as mono-spaced characters, one character per cell, with a best fitting
      font and size. For further information, see the section on
      Recommendations below. The data characters are taken sequentially from
      the data portion of the block and rendered within a cell, starting with
      the upper left cell in the grid and proceeding from left to right and
      from top to bottom. Text data is rendered until the end of data is
      reached or the character grid is filled.  The Character Grid contains an
      integral number of cells; in the case that the cell dimensions do not
      allow for an integral number, fractional cells must be discarded; an
      encoder must be careful to specify the grid dimensions accurately so that
      this does not happen. This block requires a Global Color Table to be
      available; the colors used by this block reference the Global Color Table
      in the Stream if there is one, or the Global Color Table from a previous
      Stream, if one was saved. This block is a graphic rendering block,
      therefore it may be modified by a Graphic Control Extension.  This block
      is OPTIONAL; any number of them may appear in the Data Stream.
........................
           ii) Plain Text Label - Identifies the current block as a Plain Text
            Extension. This field contains the fixed value 0x01.

            iii) Block Size - Number of bytes in the extension, after the Block
            Size field and up to but not including the beginning of the data
            portion. This field contains the fixed value 12.

            iv) Text Grid Left Position - Column number, in pixels, of the left
            edge of the text grid, with respect to the left edge of the Logical
            Screen.

            v) Text Grid Top Position - Row number, in pixels, of the top edge
            of the text grid, with respect to the top edge of the Logical
            Screen.

            vi) Image Grid Width - Width of the text grid in pixels.

            vii) Image Grid Height - Height of the text grid in pixels.

            viii) Character Cell Width - Width, in pixels, of each cell in the
            grid.

            ix) Character Cell Height - Height, in pixels, of each cell in the
            grid.

            x) Text Foreground Color Index - Index into the Global Color Table
            to be used to render the text foreground.

            xi) Text Background Color Index - Index into the Global Color Table
            to be used to render the text background.

            xii) Plain Text Data - Sequence of sub-blocks, each of size at most
            255 bytes and at least 1 byte, with the size in a byte preceding
            the data.  The end of the sequence is marked by the Block
            Terminator.

            xiii) Block Terminator - This zero-length data block marks the end
            of the Plain Text Data Blocks.

      d. Extensions and Scope. The scope of this block is the Plain Text Data
      Block contained in it. This block may be modified by the Graphic Control
      Extension.

      e. Recommendations. The data in the Plain Text Extension is assumed to be
      preformatted. The selection of font and size is left to the discretion of
      the decoder.  If characters less than 0x20 or greater than 0xf7 are
      encountered, it is recommended that the decoder display a Space character
      (0x20). The encoder should use grid and cell dimensions such that an
      integral number of cells fit in the grid both horizontally as well as
      vertically.  For broadest compatibility, character cell dimensions should
      be around 8x8 or 8x16 (width x height); consider an image for unusual
      sized text.
"! 
token
    " Private - Returns the receiver's token. "

    ^16rF9! 
blockSize
    " Returns the receiver's fixed block size in bytes. "

    ^4!   
subBlockClass
    " Private - Returns the support for sub-blocks."

    ^GIFGraphicControlSubBlock!
token
    " Private - Returns the receiver's token.
    This class can't handle any particular token. "

    ^nil!
autenticationCode
    "Returns the instance var of the receiver
     for external manipulation"
    ^autenticationCode!
autenticationCode: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    autenticationCode := anObject!   
identifier
    "Returns the instance var of the receiver
     for external manipulation"
    ^identifier!  
identifier: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    identifier := anObject! 
getContents: bytes
    " Private - Fill the receiver contents with bytes."

    | aStream |
    super getContents: bytes.
    aStream := ReadWriteStream on: bytes.
    self
        identifier: (aStream next: 8);
        autenticationCode: (aStream next: 3);
        yourself.! 
isUnknown
    "True if the receiver is unknown."

    ^true!   
isLoop
    "Returns true for a Loop block"

    ^true! 
iterations: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    iterations := anObject! 
iterations
    "Returns the instance var of the receiver
     for external manipulation"
    ^iterations!  
fileInFrom: aStream
    " Private - Fill the receiver's contents from aStream. "

    super fileInFrom: aStream.
    self
        iterations: aStream nextUShort;
        yourself! 
isComment
    "True if the receiver is aComment."

    ^true!  
pixelAspectRatio
    "Returns the instance var of the receiver
     for external manipulation"
    ^pixelAspectRatio!  
pixelAspectRatio: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    pixelAspectRatio := anObject! 
getExtrasFrom: aStream
    " Private - Get the receiver's extra information color from aStream."

    | byte |
    byte := aStream nextByte.
    self pixelAspectRatio: byte.
    ^byte!
transparentColorIndex
    "Returns the instance var of the receiver
     for external manipulation"
    ^transparentColorIndex!
transparentColorIndex: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    transparentColorIndex := anObject!   
needsUserInput
    "Returns the instance var of the receiver
     for external manipulation"
    ^needsUserInput!  
getContents: bytes
    " Private - Fill the receiver contents with bytes."

    | aStream |
    super getContents: bytes.
    aStream := ReadWriteStream on: bytes.
    self
        getFlagsFrom: aStream;
        delayTime: aStream nextUShort;
        transparentColorIndex: aStream nextByte;
        yourself.
    self useTransparentColor ifFalse: [
        self transparentColorIndex: nil
    ].!
disposeSelector
    " Returns aSymbol acording to disposal method (or nil). "

    (disposalMethod isNil or: [ disposalMethod = 0 ]) ifTrue: [ ^nil ].
    disposalMethod = 1 ifTrue: [ ^#none ]. "the graphic is to be left in place."
    disposalMethod = 2 ifTrue: [ ^#background ]. "the graphic must be restored to background."
    disposalMethod = 3 ifTrue: [ ^#previous ]. "the graphic must be restored to previous rendering."
    ^#undefined "to be defined (unknown)"
    !   
reserved
    "Returns the instance var of the receiver
     for external manipulation"
    ^reserved!  
delayTime
    "Returns the instance var of the receiver
     for external manipulation"
    ^delayTime!
delayTime: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    delayTime := anObject!   
disposalMethod
    "Returns the instance var of the receiver
     for external manipulation"
    ^disposalMethod!  
reserved: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    reserved := anObject! 
useTransparentColor
    "Returns the instance var of the receiver
     for external manipulation"
    ^useTransparentColor!
useTransparentColor: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    useTransparentColor := anObject!   
needsUserInput: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    needsUserInput := anObject! 
getFlagsFrom: aStream
    " Private - Get the receiver's flags from aStream."

    | flags |
    flags := aStream nextByte.
    self
        reserved: ((flags bitAnd: 2r11100000) bitShift: -5) "+ 1";
        disposalMethod: ((flags bitAnd: 2r00011100) bitShift: -2) + 1;
        needsUserInput: (flags bitAnd: 2r00000010) ~= 0;
        useTransparentColor: (flags bitAnd: 2r00000001) ~= 0;
        yourself.
    ^flags!
disposalMethod: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    disposalMethod := anObject! 
foreColorIndex: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    foreColorIndex := anObject! 
getExtentFrom: aStream
    " Private - Get the receiver's extent from aStream."

    | width height |
    width := aStream nextUShort.
    height := aStream nextUShort.
    ^self extent: width @ height!  
getContents: bytes
    " Private - Fill the receiver contents with bytes."

    | aStream |
    super getContents: bytes.
    aStream := ReadWriteStream on: bytes.
    self
        getLeftTopFrom: aStream;
        getExtentFrom: aStream;
        getCellExtentFrom: aStream;
        foreColorIndex: aStream nextByte;
        backColorIndex: aStream nextByte;
        yourself.!  
extent
    "Returns the instance var of the receiver
     for external manipulation"
    ^extent!  
cellExtent
    "Returns the instance var of the receiver
     for external manipulation"
    ^cellExtent!  
getCellExtentFrom: aStream
    " Private - Get the receiver's cellExtent from aStream."

    | width height |
    width := aStream nextByte.
    height := aStream nextByte.
    ^self cellExtent: width @ height!  
extent: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    extent := anObject! 
getLeftTopFrom: aStream
    " Private - Get the receiver's leftTop from aStream. "

    | left top |
    left := aStream nextUShort.
    top := aStream nextUShort.
    ^self leftTop: left @ top!  
backColorIndex
    "Returns the instance var of the receiver
     for external manipulation"
    ^backColorIndex!  
cellExtent: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    cellExtent := anObject! 
foreColorIndex
    "Returns the instance var of the receiver
     for external manipulation"
    ^foreColorIndex!  
leftTop
    "Returns the instance var of the receiver
     for external manipulation"
    ^leftTop!
leftTop: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    leftTop := anObject!   
backColorIndex: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    backColorIndex := anObject! 
next
    " Get the next element from receiver at current location. "

    ^(self species blockClass
        for: self stream nextByte in: self)
        fileInFrom: self stream! 
position
    " Returns the stream position. "

    ^self stream position!  
signature
    "Returns the instance var of the receiver
     for external manipulation"
    ^signature!
signature: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    signature := anObject!   
colorMap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    colorMap := anObject! 
stream
    "Returns the instance var of the receiver
     for external manipulation"
    ^stream!  
startPosition
    "Returns the instance var of the receiver
     for external manipulation"
    ^startPosition!
startPosition: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    startPosition := anObject!   
peek
    " Peek the next element from the receiver. "

    | position element |
    self atEnd ifTrue: [ ^nil ].
    position := self position.
    element := self next.
    self position: position.
    ^element ! 
extent
    "Returns the screen extent."

    ^self screenDescriptor extent!
getColorMap
    " Private - Get the receiver's GlobalColorMap information from current stream. "

    | map |
    self screenDescriptor useColorMap ifFalse: [
        ^self colorMap: nil
    ].
    map := self species colorMapClass
        bitsPerPixel: self screenDescriptor bitsPerPixel
        from: self stream.
    self colorMap: map.!
getScreenDescriptor
    " Private - Get the receiver's ScreenDescriptor information from current stream. "

    self screenDescriptor: self species screenDescriptorClass new.
    self screenDescriptor fileInFrom: self stream.!
getHeader
    " Private - Get the receiver's header information from current stream. "

    self
        getSignature;
        getScreenDescriptor;
        getColorMap;
        yourself! 
screenDescriptor: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    screenDescriptor := anObject! 
position: aPosition
    " Sets the receiver at aPosition. "

    ^self stream position: aPosition! 
getSignature
    " Private - Get the receiver's signature information from current stream. "

    signature := self stream next: 6.!   
upTo: element
    " Returns the elements in receiver from current location up to element. "

    | result current |
    result := OrderedCollection new.
    [self atEnd] whileFalse: [
        current := self next.
        current = element ifTrue: [ ^result asArray ].
        result add: current.
    ].
    ^result asArray!   
stream: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    stream := anObject! 
readFrom: aStream
    " Private - Set the receiver to read from aStream. "

    self
        stream: aStream;
        getHeader.
    self startPosition: self position.!
next: count
    " Get the next <count> elements from receiver at current location. "

    | result |
    result := OrderedCollection new.
    count timesRepeat: [
        self atEnd ifFalse: [ result add: self next ]
    ].
    ^result!  
close
    "Close the receiver's stream."

    self stream close.!  
colorMap
    "Returns the instance var of the receiver
     for external manipulation"
    ^colorMap!  
contents
    "Returns the receiver's contents."

    ^self reset; upTo: nil!   
atEnd
    "True if the receiver is atEnd."

    ^self stream atEnd!
screenDescriptor
    "Returns the instance var of the receiver
     for external manipulation"
    ^screenDescriptor!  
reset
    " Reset the receiver to the first element position. "

    self position: self startPosition.!   
adaptContent: bytes
    " Private - The receiver's can adapt bytes contents to internal representation."

    | aClass |
    aClass := bytes = self contents first
        ifTrue: [ GIFGraphicControlSubBlock ]
        ifFalse: [ GIFDataSubBlock ].
    ^aClass
        in: self
        with: bytes! 
isControl
    "True if the receiver is aControl."

    ^true!  
version
    " Private - Must return the receiver's version string."

    ^'87a'!   
screenDescriptorClass
    " Private - Must return the class for supoprt of screen descriptor."

    ^GIF87aScreenDescriptor!   
canHandle: aGIFObject
    " Must returns true when the receiver can wrap aGIFObject."

    ^aGIFObject isUnknown!